from typing import Any


# ------------------ Base ------------------- #
class BaseInpostError(Exception):
    """Base exception to inherit from

    :param reason: reason of BaseInpostError happening
    :type reason: typing.Any"""

    def __init__(self, reason):
        """Constructor method

        :param reason: Reason of error
        :type reason: Any
        """
        super().__init__(reason)
        self.reason: Any = reason

    @property
    def stacktrace(self) -> Any:
        """Gets stacktrace of raised exception
        :return: reason why exception occured
        :rtype: Any"""
        return self.reason


# ----------------- Parcels ----------------- #


class ParcelTypeError(BaseInpostError):
    """Is raised when expected :class:`ParcelType` does not match with actual one"""

    pass


class NoParcelError(BaseInpostError):
    """Is raised when no parcel is set in :class:`Parcel`"""

    pass


class UnknownStatusError(BaseInpostError):
    """Is raised when no status matches"""

    pass


class UnidentifiedParcelError(BaseInpostError):
    """Is raised when no other :class:`Parcel` error match"""

    pass


# ----------------- API ----------------- #
class NotAuthenticatedError(BaseInpostError):
    """Is raised when `Inpost.auth_token` is missing"""

    pass


class ReAuthenticationError(BaseInpostError):
    """Is raised when `Inpost.auth_token` has expired"""

    pass


class PhoneNumberError(BaseInpostError):
    """Is raised when `Inpost.phone_number` is invalid or unexpected error connected with phone number occurs"""

    pass


class SmsCodeError(BaseInpostError):
    """Is raised when `Inpost.sms_code` is invalid or unexpected sms_code occurs"""

    pass


class RefreshTokenError(BaseInpostError):
    """Is raised when `Inpost.refr_token` is invalid or unexpected error connected with refresh token occurs"""

    pass


class NotFoundError(BaseInpostError):
    """Is raised when method from :class:`Inpost` returns 404 Not Found HTTP status code"""

    pass


class UnauthorizedError(BaseInpostError):
    """Is raised when method from :class:`Inpost` returns 401 Unauthorized HTTP status code"""

    pass


class UnidentifiedAPIError(BaseInpostError):
    """Is raised when no other API error match"""

    pass


# ----------------- Other ----------------- #
class UserLocationError(BaseInpostError):
    pass


class SingleParamError(BaseInpostError):
    """Is raised when only one param must be filled in but got more"""

    pass


class MissingParamsError(BaseInpostError):
    """Is raised when none of params are filled"""

    pass


class UnidentifiedError(BaseInpostError):
    """Is raised when no other error match"""

    pass



import logging

from arrow import Arrow, get


class Notification:
    """Object representation of :class:`inpost.api.Inpost` notification

    :param notification_data: :class:`dict` containing all notification data
    :type notification_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, notification_data: dict, logger: logging.Logger):
        """Constructor method

        :param notification_data: :class:`dict` containing all notification data
        :type notification_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.id: str = notification_data.get("id", None)
        self._log: logging.Logger = logger.getChild(f"{self.__class__.__name__}.{self.id}")
        self.type: str = notification_data.get("type", None)
        self.action: str = notification_data.get("action", None)
        self.date: Arrow = get(notification_data.get("date")) if "date" in notification_data else None
        self.title: str = notification_data.get("title", None)
        self.content: str = notification_data.get("content", None)
        self.shipment_number: str = notification_data.get("shipmentNumber", None)
        self.read: bool = notification_data.get("read", None)
        self.extra_params: dict = notification_data.get("extraParams", None)
        self.parcel_type: str = notification_data.get("parcelType", None)

        self._log.debug(f"created notification with id {self.id}")



import logging

from arrow import Arrow, get


class Friend:
    """Object representation of :class:`inpost.api.Inpost` friend

    :param friend_data: :class:`dict` containing all friend data
    :type friend_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, friend_data: dict, logger: logging.Logger):
        """Constructor method

        :param friend_data: :class:`dict` containing all friend data
        :type friend_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.uuid: str = friend_data["uuid"] if "uuid" in friend_data else None
        self.phone_number: str = friend_data["phoneNumber"]
        self.name: str = friend_data["name"]
        self._log: logging.Logger = logger.getChild(f"{self.__class__.__name__}.{self.uuid}")
        self.invitaion_code: str | None = friend_data["invitationCode"] if "invitationCode" in friend_data else None
        self.created_date: Arrow | None = get(friend_data["createdDate"]) if "createdDate" in friend_data else None
        self.expiry_date: Arrow | None = get(friend_data["expiryDate"]) if "expiryDate" in friend_data else None

        if self.invitaion_code:
            self._log.debug(f"created friendship with {self.name} using from_invitation")
        else:
            self._log.debug(f"created friendship with {self.name}")

    @classmethod
    def from_invitation(cls, invitation_data: dict, logger: logging.Logger):
        """`Classmethod` to initialize :class:`Friend` from incitation.
        Should be used when retrieving configuration from database.

        :param invitation_data: :class:`dict` containing all friend data
        :type invitation_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        :return: Friend object from provided invitation
        :rtype: Friend
        """

        return cls(
            friend_data={
                "uuid": invitation_data["friend"]["uuid"],
                "phoneNumber": invitation_data["friend"]["phoneNumber"],
                "name": invitation_data["friend"]["name"],
                "invitationCode": invitation_data["invitationCode"],
                "createdDate": invitation_data["createdDate"],
                "expiryDate": invitation_data["expiryDate"],
            },
            logger=logger,
        )

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")



# AUTH #
login_url: str = "https://api-inmobile-pl.easypack24.net/v1/authenticate"
send_sms_code_url: str = "https://api-inmobile-pl.easypack24.net/v1/sendSMSCode/"  # get
confirm_sms_code_url: str = "https://api-inmobile-pl.easypack24.net/v1/confirmSMSCode"  # post
logout_url: str = "https://api-inmobile-pl.easypack24.net/v1/logout"  # post
refresh_token_url: str = "https://api-inmobile-pl.easypack24.net/v1/authenticate"  # post

# INCOMING PARCELS #
tracked_url: str = "https://api-inmobile-pl.easypack24.net/v3/parcels/tracked/"  # get
multi_url: str = "https://api-inmobile-pl.easypack24.net/v3/parcels/multi/"  # get
collect_url: str = "https://api-inmobile-pl.easypack24.net/v1/collect/validate"  # post
compartment_reopen_url: str = "https://api-inmobile-pl.easypack24.net/v1/collect/compartment/reopen"  # post
compartment_open_url: str = "https://api-inmobile-pl.easypack24.net/v1/collect/compartment/open"  # post
compartment_status_url: str = "https://api-inmobile-pl.easypack24.net/v1/collect/compartment/status"  # post
terminate_collect_session_url: str = "https://api-inmobile-pl.easypack24.net/v1/collect/terminate"  # post
shared_url: str = "https://api-inmobile-pl.easypack24.net/v1/parcels/shared"  # post

# CREATING PARCEL #
create_url: str = "https://api-inmobile-pl.easypack24.net/v1/parcels"
points_url: str = "https://api-inmobile-pl.easypack24.net/v3/points"
blik_status_url: str = "https://api-inmobile-pl.easypack24.net/v1/payments/blik/alias/status"
create_blik_url: str = "https://api-inmobile-pl.easypack24.net/v1/payments/transactions/create/blik"

# OUTGOING PARCELS #
sent_url: str = "https://api-inmobile-pl.easypack24.net/v2/parcels/sent/"  # get
parcel_points_url: str = "https://api-inmobile-pl.easypack24.net/v3/points/"  # get
validate_sent_url: str = "https://api-inmobile-pl.easypack24.net/v1/send/validate/"  # post
open_sent_url: str = "https://api-inmobile-pl.easypack24.net/v1/send/compartment/open"  # post
reopen_sent_url: str = "https://api-inmobile-pl.easypack24.net/v1/send/compartment/reopen"  # post
status_sent_url: str = "https://api-inmobile-pl.easypack24.net/v1/send/compartment/status"  # post
confirm_sent_url: str = "https://api-inmobile-pl.easypack24.net/v1/send/confirm"  # post
parcel_prices_url: str = "https://api-inmobile-pl.easypack24.net/v1/prices/parcels"  # get

# RETURNS #
returns_url: str = "https://api-inmobile-pl.easypack24.net/v1/returns/parcels/"  # get
tickets_url: str = "https://api-inmobile-pl.easypack24.net/v1/returns/tickets"  # get
parcel_notifications_url: str = "https://api-inmobile-pl.easypack24.net/v2/notifications?type=PUSH%2CNEWS%2CTILE"  # get

# FRIENDS #
friendship_url: str = "https://api-inmobile-pl.easypack24.net/v1/friends/"  # get, post, patch, delete
validate_friendship_url: str = "https://api-inmobile-pl.easypack24.net/v1/invitations/validate"  # post
accept_friendship_url: str = "https://api-inmobile-pl.easypack24.net/v1/invitations/accept"  # post



import logging
from typing import List

from aiohttp import ClientResponse, ClientSession
from aiohttp.typedefs import StrOrURL

from inpost.static import (
    CompartmentExpectedStatus,
    DeliveryType,
    Friend,
    MissingParamsError,
    NoParcelError,
    NotAuthenticatedError,
    NotFoundError,
    Parcel,
    ParcelCarrierSize,
    ParcelLockerSize,
    ParcelPointOperations,
    ParcelShipmentType,
    ParcelStatus,
    ParcelType,
    ParcelTypeError,
    PhoneNumberError,
    Point,
    ReAuthenticationError,
    Receiver,
    RefreshTokenError,
    ReturnParcel,
    Sender,
    SentParcel,
    SingleParamError,
    SmsCodeError,
    UnauthorizedError,
    UnidentifiedAPIError,
    appjson,
    blik_status_url,
    collect_url,
    compartment_open_url,
    compartment_reopen_url,
    compartment_status_url,
    confirm_sms_code_url,
    create_blik_url,
    create_url,
    friendship_url,
    logout_url,
    multi_url,
    open_sent_url,
    parcel_points_url,
    parcel_prices_url,
    refresh_token_url,
    returns_url,
    send_sms_code_url,
    sent_url,
    shared_url,
    status_sent_url,
    terminate_collect_session_url,
    tracked_url,
    validate_friendship_url,
    validate_sent_url,
)


class Inpost:
    """Python representation of an Inpost app. Essentially implements methods to manage all incoming parcels"""

    def __init__(self, phone_number):
        """Constructor method

        :param phone_number: phone number
        :type phone_number: str
        :raises PhoneNumberError: Wrong phone number format or is not digit
        """

        if isinstance(phone_number, int):
            phone_number = str(phone_number)

        if not (len(phone_number) == 9 and phone_number.isdigit()):
            raise PhoneNumberError(f"Wrong phone number format: {phone_number} (should be 9 digits)")

        self.phone_number: str = phone_number
        self.sms_code: str | None = None
        self.auth_token: str | None = None
        self.refr_token: str | None = None
        self.sess: ClientSession = ClientSession()
        self._log = logging.getLogger(f"{self.__class__.__name__}.{phone_number}")

        self._log.setLevel(level=logging.DEBUG)
        self._log.info(f"initialized inpost object with phone number {phone_number}")

    def __repr__(self):
        return f"{self.__class__.__name__}(phone_number={self.phone_number})"

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return self.disconnect()

    async def request(
        self,
        method: str,
        action: str,
        url: StrOrURL,
        auth: bool = True,
        headers: dict | None = None,
        params: dict | None = None,
        data: dict | None = None,
        autorefresh: bool = True,
        **kwargs,
    ) -> ClientResponse:
        """Validates sent data and fetches required compartment properties for opening

        :param method: HTTP method of request
        :type method: str
        :param action: action type (e.g. "get parcels" or "send sms code") for logging purposes
        :type action: str
        :param url: HTTP request url
        :type url: StrOrURL
        :param auth: True if request should contain authorization header else False
        :type auth: bool
        :param headers: Additional headers for HTTP request (don't put authorization header here)
        :type headers: dict | None
        :param params: dict of parameters to get method
        :type params: dict | None
        :param data: Data to be sent in HTTP request
        :type data: dict | None
        :param autorefresh: method automatically try to refresh token if API returns HTTP 401 Unauthorized status code
        :type autorefresh: bool
        :param kwargs: additional keyword arguments
        :return: response of http request
        :rtype: ClientResponse
        :raises UnauthorizedError: User not authenticated in inpost service
        :raises NotFoundError: URL not found
        :raises UnidentifiedAPIError: Unexpected things happened
        :raises ValueError: Doubled authorization header in request
        """

        if auth and headers:
            if "Authorization" in headers:
                raise ValueError("Both auth==True and Authorization in additional headers")

        headers_ = {} if headers is None else headers

        if auth:
            headers_.update({"Authorization": self.auth_token})

        resp = await self.sess.request(method, url, headers=headers_, params=params, json=data, **kwargs)

        if autorefresh and resp.status == 401:
            await self.refresh_token()
            headers_.update({"Authorization": self.auth_token})
            resp = await self.sess.request(method, url, headers=headers_, params=params, json=data, **kwargs)

        match resp.status:
            case 200:
                self._log.debug(f"{action} done")
                return resp
            case 401:
                self._log.error(f"could not perform {action}, unauthorized")
                raise UnauthorizedError(reason=resp)
            case 404:
                self._log.error(f"could not perform {action}, not found")
                raise NotFoundError(reason=resp)
            case _:
                self._log.error(f"could not perform {action}, unhandled status")

        raise UnidentifiedAPIError(reason=resp)

    @classmethod
    def from_dict(cls, data: dict) -> "Inpost":
        """`Classmethod` to initialize :class:`Inpost` object with dict.
        Should be used when retrieving configuration from database.

        :param data: User's Inpost data (e.g. phone_number, sms_code, auth_token, refr_token)
        :type data: dict
        :return: Inpost object from provided dict
        :rtype: Inpost
        """

        inp = cls(phone_number=data["phone_number"])
        inp.sms_code = data["sms_code"]
        inp.auth_token = data["auth_token"]
        inp.refr_token = data["refr_token"]

        inp._log.info("initialized by from_dict")
        return inp

    async def send_sms_code(self) -> bool:
        """Sends sms code to `Inpost.phone_number`

        :return: True if sms code is sent else False
        :rtype: bool
        :raises PhoneNumberError: Missing phone number
        """

        if not self.phone_number:  # can't log it cuz if there's no phone number no logger initialized @shrug
            raise PhoneNumberError("Phone number missing")

        self._log.info("sending sms code")

        resp = await self.request(
            method="post",
            action="send sms code",
            url=send_sms_code_url,
            auth=False,
            headers=None,
            data={"phoneNumber": f"{self.phone_number}"},
            autorefresh=False,
        )

        return resp.status == 200

    async def confirm_sms_code(self, sms_code: str | int) -> bool:
        """Confirms sms code sent to `Inpost.phone_number` and fetches tokens

        :param sms_code: sms code sent to Inpost.phone_number device
        :type sms_code: str | int
        :return: True if sms code gets confirmed and tokens fetched
        :rtype: bool
        :raises PhoneNumberError: Missing phone number
        :raises SmsCodeError: Wrong sms code format
        """

        if not self.phone_number:  # can't log it cuz if there's no phone number no logger initialized @shrug
            raise PhoneNumberError("Phone number missing")

        if isinstance(sms_code, int):
            sms_code = str(sms_code)

        if len(sms_code) != 6 or not sms_code.isdigit():
            raise SmsCodeError(reason=f"Wrong sms code format: {sms_code} (should be 6 digits)")

        self._log.info("confirming sms code")

        resp = await self.request(
            method="post",
            action="confirm sms code",
            url=confirm_sms_code_url,
            auth=False,
            headers=appjson,
            data={"phoneNumber": self.phone_number, "smsCode": sms_code, "phoneOS": "Android"},
            autorefresh=False,
        )

        if resp.status == 200:
            auth_token_data = await resp.json()
            self.sms_code = sms_code
            self.refr_token = auth_token_data["refreshToken"]
            self.auth_token = auth_token_data["authToken"]
            self._log.debug("sms code confirmed")
            return True

        return False

    async def refresh_token(self) -> bool:
        """Refreshes authorization token using refresh token

        :return: True if Inpost.auth_token gets refreshed
        :rtype: bool
        :raises RefreshTokenError: Missing refresh token
        :raises ReAuthenticationError: Re-authentication needed
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info("refreshing token")

        if not self.refr_token:
            self._log.error("refresh token missing")
            raise RefreshTokenError(reason="Refresh token missing")

        resp = await self.request(
            method="post",
            action="refresh token",
            url=refresh_token_url,
            auth=False,
            headers=appjson,
            data={"refreshToken": self.refr_token, "phoneOS": "Android"},
            autorefresh=False,
        )

        if resp.status == 200:
            confirmation = await resp.json()
            if confirmation["reauthenticationRequired"]:
                self._log.error("could not refresh token, log in again")
                raise ReAuthenticationError(reason="You need to log in again!")

            self.auth_token = confirmation["authToken"]
            self._log.debug("token refreshed")
            return True

        return False

    async def logout(self) -> bool:
        """Logouts user from inpost api service

        :return: True if the user is logged out
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info("logging out")

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post", action="logout", url=logout_url, auth=True, headers=None, data=None, autorefresh=True
        )

        if resp.status == 200:
            self.phone_number = ""
            self.refr_token = None
            self.auth_token = None
            self.sms_code = None
            self._log.debug("logged out")
            return True

        return False

    async def disconnect(self) -> bool:
        """Simplified method to logout and close user's session

        :return: True if user is logged out and session is closed else False
        :raises NotAuthenticatedError: User not authenticated in inpost service
        """

        self._log.info("disconnecting")
        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        if await self.logout():
            await self.sess.close()
            self._log.debug("disconnected")
            return True

        self._log.error("could not disconnect")
        return False

    async def get_parcel(
        self, shipment_number: int | str, parcel_type: ParcelType = ParcelType.TRACKED, parse=False
    ) -> dict | Parcel | SentParcel | ReturnParcel:
        """Fetches single parcel from provided shipment number

        :param shipment_number: Parcel's shipment number
        :type shipment_number: int | str
        :param parcel_type: Parcel type (e.g. received, sent, returned)
        :type parcel_type: ParcelType
        :param parse: if set to True method will return :class:`Parcel` else :class:`dict`
        :type parse: bool
        :return: Fetched parcel data
        :rtype: dict | Parcel
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        :raises ParcelTypeError: Unknown parcel type selected
        """

        self._log.info(f"getting parcel with shipment number: {shipment_number}")

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        match parcel_type:
            case ParcelType.TRACKED:
                self._log.debug(f"getting parcel type {parcel_type}")
                url = tracked_url
            case ParcelType.SENT:
                self._log.debug(f"getting parcel type {parcel_type}")
                url = sent_url
            case ParcelType.RETURNS:
                self._log.debug(f"getting parcel type {parcel_type}")
                url = returns_url
            case _:
                self._log.error(f"unexpected parcel type {parcel_type}")
                raise ParcelTypeError(reason=f"Unexpected parcel type: {parcel_type}")

        resp = await self.request(
            method="get",
            action=f"parcel with shipment number {shipment_number}",
            url=f"{url}{shipment_number}",
            auth=True,
            headers=None,
            data=None,
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"parcel with shipment number {shipment_number} received")
            match parcel_type:
                case ParcelType.TRACKED:
                    return await resp.json() if not parse else Parcel(await resp.json(), logger=self._log)
                case ParcelType.SENT:
                    return await resp.json() if not parse else SentParcel(await resp.json(), logger=self._log)
                case ParcelType.RETURNS:
                    return await resp.json() if not parse else ReturnParcel(await resp.json(), logger=self._log)
                case _:
                    self._log.error(f"wrong parcel type {parcel_type}")
                    raise ParcelTypeError(reason=f"Unknown parcel type: {parcel_type}")

        raise UnidentifiedAPIError(reason=resp)

    async def get_parcels(
        self,
        parcel_type: ParcelType = ParcelType.TRACKED,
        status: ParcelStatus | List[ParcelStatus] | None = None,
        pickup_point: str | List[str] | None = None,
        shipment_type: ParcelShipmentType | List[ParcelShipmentType] | None = None,
        parse: bool = False,
    ) -> List[dict] | List[Parcel]:
        """Fetches all available parcels for set `Inpost.phone_number` and optionally filters them

        :param parcel_type: Parcel type (e.g. received, sent, returned)
        :type parcel_type: ParcelType
        :param status: status that each fetched parcels has to be in
        :type status: ParcelStatus | list[ParcelStatus] | None
        :param pickup_point: Fetched parcels have to be picked from this pickup point (e.g. `GXO05M`)
        :type pickup_point: str | list[str] | None
        :param shipment_type: Fetched parcels have to be shipped that way
        :type shipment_type: ParcelShipmentType | list[ParcelShipmentType] | None
        :param parse: if set to True method will return list[:class:`Parcel`] else list[:class:`dict`]
        :type parse: bool
        :return: fetched parcels data
        :rtype: list[dict] | list[Parcel]
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises ParcelTypeError: Unknown parcel type selected
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info("getting parcels")

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        match parcel_type:
            case ParcelType.TRACKED:
                self._log.debug(f"getting parcel type {parcel_type}")
                url = tracked_url
            case ParcelType.SENT:
                self._log.debug(f"getting parcel type {parcel_type}")
                url = sent_url
            case ParcelType.RETURNS:
                self._log.debug(f"getting parcel type {parcel_type}")
                url = returns_url
            case _:
                self._log.error(f"wrong parcel type {parcel_type}")
                raise ParcelTypeError(reason=f"Unknown parcel type: {parcel_type}")

        async with await self.request(
            method="get", action="get parcels", url=url, auth=True, headers=None, data=None, autorefresh=True
        ) as resp:
            if resp.status != 200:
                self._log.debug(f"Could not get parcels due to HTTP error {resp.status}")
                raise UnidentifiedAPIError(reason=resp)

            self._log.debug(f"received {parcel_type} parcels")
            _parcels = (await resp.json())["parcels"]

            if status is not None:
                if isinstance(status, ParcelStatus):
                    status = [status]

                _parcels = (_parcel for _parcel in _parcels if ParcelStatus[_parcel.get("status")] in status)

            if pickup_point is not None:
                if isinstance(pickup_point, str):
                    pickup_point = [pickup_point]

                _parcels = (_parcel for _parcel in _parcels if _parcel["pickUpPoint"]["name"] in pickup_point)

            if shipment_type is not None:
                if isinstance(shipment_type, ParcelShipmentType):
                    shipment_type = [shipment_type]

                _parcels = (
                    _parcel
                    for _parcel in _parcels
                    if ParcelShipmentType[_parcel["shipmentType"]] in list(shipment_type)
                )

            return list(_parcels) if not parse else [Parcel(parcel_data=data, logger=self._log) for data in _parcels]

    async def get_multi_compartment(self, multi_uuid: str | int, parse: bool = False) -> dict | List[Parcel]:
        """Fetches all available parcels for set `Inpost.phone_number` and optionally filters them

        :param multi_uuid: multicompartment uuid
        :type multi_uuid: str | int
        :param parse: switch for parsing response
        :type parse: bool
        :return: multicompartment parcels
        :rtype: dict | List[Parcel]
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="get",
            action=f"parcel with multi-compartment uuid {multi_uuid}",
            url=f"{multi_url}{multi_uuid}",
            auth=True,
            headers=None,
            data=None,
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"parcel with multi-compartment uuid {multi_uuid} received")
            return (
                (await resp.json())["parcels"]
                if not parse
                else [Parcel(data, logger=self._log) for data in (await resp.json())["parcels"]]
            )

        raise UnidentifiedAPIError(reason=resp)

    async def collect_compartment_properties(
        self, shipment_number: str | int | None = None, parcel_obj: Parcel | None = None, location: dict | None = None
    ) -> Parcel:
        """Validates sent data and fetches required compartment properties for opening

        :param shipment_number: Parcel's shipment number
        :type shipment_number: int | str | None
        :param parcel_obj: :class:`Parcel` object to obtain data from
        :type parcel_obj: Parcel | None
        :param location: Fetched parcels have to be picked from this pickup point (e.g. `GXO05M`)
        :type location: dict | None
        :return: fetched parcels data
        :rtype: Parcel
        :raises MissingParamsError: none of required query and location params are filled
        :raises SingleParamError: Fields shipment_number and parcel_obj filled in but only one of them is required
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises NoParcelError: Could not get parcel object from provided data
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened

        .. warning:: you must fill in only one parameter - shipment_number or parcel_obj!
        """

        parcel_obj_: Parcel | None = None
        if shipment_number is None is parcel_obj:
            self._log.error("none of shipment_number and parcel_obj filled in")
            raise MissingParamsError(reason="None of params are filled (one required)")
        elif shipment_number is not None is not parcel_obj:
            self._log.error("shipment_number and parcel_obj filled in")
            raise SingleParamError(reason="Fields shipment_number and parcel_obj filled in! Choose one!")
        elif shipment_number:
            self._log.debug(f"parcel_obj not provided, getting from shipment number {shipment_number}")
            parcel_obj_ = await self.get_parcel(shipment_number=shipment_number, parse=True)
        elif parcel_obj:
            parcel_obj_ = parcel_obj

        if parcel_obj_ is None:
            raise NoParcelError(reason="Could not obtain desired parcel!")

        self._log.info(f"collecting compartment properties for {parcel_obj_.shipment_number}")

        resp = await self.request(
            method="post",
            action="collect compartment properties",
            url=collect_url,
            auth=True,
            headers=None,
            data={
                "parcel": parcel_obj_.compartment_open_data,
                "geoPoint": location if location is not None else parcel_obj_.mocked_location,
            },
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"collected compartment properties for {parcel_obj_.shipment_number}")
            parcel_obj_.compartment_properties = await resp.json()
            return parcel_obj_

        raise UnidentifiedAPIError(reason=resp)

    async def open_compartment(self, parcel_obj: Parcel) -> Parcel:
        """Opens compartment for `Inpost.parcel` object

        :param parcel_obj: Parcel object
        :type parcel_obj: Parcel
        :return: Parcel with compartment location set if it gets opened
        :rtype: Parcel
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info(f"opening compartment for {parcel_obj.shipment_number}")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post",
            action=f"open compartment for {parcel_obj.shipment_number}",
            url=compartment_open_url,
            auth=True,
            headers=None,
            data={"sessionUuid": parcel_obj.compartment_properties.session_uuid},
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"opened compartment for {parcel_obj.shipment_number}")
            parcel_obj.compartment_location = await resp.json()
            return parcel_obj

        raise UnidentifiedAPIError(reason=resp)

    async def check_compartment_status(
        self, parcel_obj: Parcel, expected_status: CompartmentExpectedStatus = CompartmentExpectedStatus.OPENED
    ) -> bool:
        """Checks and compare compartment status (e.g. opened, closed) with expected status

        :param parcel_obj: Parcel object
        :type parcel_obj: Parcel
        :param expected_status: Compartment expected status
        :type expected_status: CompartmentExpectedStatus
        :return: True if actual status equals expected status else False
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info(f"checking compartment status for {parcel_obj.shipment_number}")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post",
            action="check compartment status",
            url=compartment_status_url,
            auth=True,
            headers=None,
            data={
                "sessionUuid": parcel_obj.compartment_properties.session_uuid,
                "expectedStatus": expected_status.name,
            },
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"checked compartment status for {parcel_obj.shipment_number}")
            parcel_obj.compartment_status = (await resp.json())["status"]
            return CompartmentExpectedStatus[(await resp.json())["status"]] == expected_status

        raise UnidentifiedAPIError(reason=resp)

    async def terminate_collect_session(self, parcel_obj: Parcel) -> bool:
        """Terminates user session in inpost api service

        :param parcel_obj: Parcel object
        :type parcel_obj: Parcel
        :return: True if the user session is terminated
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info(f"terminating collect session for {parcel_obj.shipment_number}")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post",
            action="terminate collect session",
            url=terminate_collect_session_url,
            auth=True,
            headers=None,
            data={"sessionUuid": parcel_obj.compartment_properties.session_uuid},
            autorefresh=True,
        )
        if resp.status == 200:
            self._log.debug(f"terminated collect session for {parcel_obj.shipment_number}")
            return True

        raise UnidentifiedAPIError(reason=resp)

    async def collect(
        self, shipment_number: str | None = None, parcel_obj: Parcel | None = None, location: dict | None = None
    ) -> Parcel | None:
        """Simplified method to open compartment

        :param shipment_number: Parcel's shipment number
        :type shipment_number: int | str | None
        :param parcel_obj: :class:`Parcel` object to obtain data from
        :type parcel_obj: Parcel | None
        :param location: Fetched parcels have to be picked from this pickup point (e.g. `GXO05M`)
        :type location: dict | None
        :return: fetched parcels data
        :rtype: bool
        :raises SingleParamError: Fields shipment_number and parcel_obj filled in but only one of them is required
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises NoParcelError: Could not get parcel object from provided data
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened

        .. warning:: you must fill in only one parameter - shipment_number or parcel_obj!
        """

        if shipment_number and parcel_obj:
            self._log.error("shipment_number and parcel_obj filled in")
            raise SingleParamError(reason="Fields shipment_number and parcel_obj filled! Choose one!")

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        if shipment_number is not None and parcel_obj is None:
            parcel_obj = await self.get_parcel(shipment_number=shipment_number, parse=True)

        if parcel_obj is None:
            raise NoParcelError(reason="Could not obtain desired parcel!")

        self._log.info(f"collecting parcel with shipment number {parcel_obj.shipment_number}")

        if parcel_obj_ := await self.collect_compartment_properties(parcel_obj=parcel_obj, location=location):
            if parcel_obj__ := await self.open_compartment(parcel_obj=parcel_obj_):
                if await self.check_compartment_status(parcel_obj=parcel_obj__):
                    return parcel_obj__

        return None

    async def close_compartment(self, parcel_obj: Parcel) -> bool:
        """Checks whether actual compartment status and expected one matches then notifies inpost api that
        compartment is closed. Should be invoked after collecting parcel

        :param parcel_obj: Parcel object
        :type parcel_obj: Parcel
        :return: True if compartment status is closed and successfully terminates user's session else False
        :rtype: bool
        """

        self._log.info(f"closing compartment for {parcel_obj.shipment_number}")

        if await self.check_compartment_status(expected_status=CompartmentExpectedStatus.CLOSED, parcel_obj=parcel_obj):
            if await self.terminate_collect_session(parcel_obj=parcel_obj):
                return True

        return False

    async def reopen_compartment(self, parcel_obj: Parcel) -> bool:
        """Reopens compartment for `Inpost.parcel` object

        :param parcel_obj: Parcel object
        :type parcel_obj: Parcel
        :return: True if compartment gets reopened
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info(f"reopening compartment for {parcel_obj.shipment_number}")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post",
            action=f"reopen compartment for {parcel_obj.shipment_number}",
            url=compartment_reopen_url,
            auth=True,
            headers=None,
            data={"sessionUuid": parcel_obj.compartment_properties.session_uuid},
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"reopened compartment for {parcel_obj.shipment_number}")
            return True

        raise UnidentifiedAPIError(reason=resp)

    async def get_parcel_points(
        self,
        query: str | None = None,
        latitude: float | None = None,
        longitude: float | None = None,
        per_page: int = 1000,
        operation: ParcelPointOperations = ParcelPointOperations.CREATE,
        parse: bool = True,
    ) -> dict | List[Point]:
        """Fetches parcel points for inpost services

        :param query: parcel point search query (e.g. GXO05M)
        :type query: str | None
        :param latitude: latitude of place we are looking for nearby parcel points
        :type latitude: float | None
        :param longitude: longitude of place we are looking for nearby parcel points
        :type longitude: float | None
        :param per_page: number of parcel points we would like to get from query, defaults to 1000
        :type per_page: int
        :param operation: operation you want to perform (e.g. CREATE, SEND)
        :type operation: ParcelPointOperations
        :param parse: parse output or not, defaults to True
        :type parse: bool
        :return: :class:`dict` of prices for inpost services
        :rtype: dict
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises SingleParamError: query and location params filled, but only one is required
        :raises MissingParamsError: none of required query and location params are filled
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info("getting parcel prices")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        if query is not None and (latitude is not None and longitude is not None):
            self._log.debug("query and location provided")
            raise SingleParamError(reason="Fields query and location filled! Chose one!")

        _params = {"filter": operation.value, "perPage": per_page}

        if query is not None:
            _params.update({"query": query})
        elif latitude is not None and longitude is not None:
            _params.update({"relative_point": f"{latitude},{longitude}"})
        else:
            raise MissingParamsError(reason="None of params are filled (one required)")

        resp = await self.request(
            method="get",
            action="get parcel points",
            url=parcel_points_url,
            auth=True,
            headers=None,
            params=_params,
            data=None,
            autorefresh=True,
        )
        if resp.status == 200:
            self._log.debug("got parcel prices")
            return (
                await resp.json()
                if not parse
                else [Point(point_data=point, logger=self._log) for point in (await resp.json())["points"]]
            )

        raise UnidentifiedAPIError(reason=resp)

    async def blik_status(self) -> bool:
        """Checks if user has active blik session


        :return: True if user has no active blik sessions else False
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        """

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        self._log.info("checking if user has opened blik session")

        resp = await self.request(
            method="get",
            action="check user blik session",
            url=blik_status_url,
            auth=True,
            headers=None,
            autorefresh=True,
        )

        if resp.status == 200 and not (await resp.json())["active"]:
            self._log.debug("user has no active blik sessions")
            return True

        return False

    async def create_parcel(
        self,
        delivery_type: DeliveryType,
        parcel_size: ParcelLockerSize | ParcelCarrierSize,
        price: float | str,
        sender: Sender,
        receiver: Receiver,
        delivery_point: Point,
    ) -> dict | None:
        """Fetches parcel points for inpost services

        :param delivery_type: a way parcel will be delivered
        :type delivery_type: DeliveryType
        :param parcel_size: size of parcel
        :type parcel_size: ParcelLockerSize | ParcelCarrierSize
        :param price: price for chosen parcel size
        :type price: float | str
        :param sender: parcel sender
        :type sender: Sender
        :param receiver: parcel receiver
        :type receiver: Receiver
        :param delivery_point: parcel delivery point
        :type delivery_point: Point
        :return: :class:`dict` with response
        :rtype: dict | None
        :raises UnidentifiedAPIError: Unexpected thing happened
        :raises NotAuthenticatedError: User not authenticated in inpost service
        """

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        self._log.info("creating new parcel")

        resp = await self.request(
            method="post",
            action="create parcel",
            url=create_url,
            auth=True,
            headers=None,
            data={
                "deliveryType": delivery_type.value,
                "parcelSize": parcel_size.name,
                "price": price,
                "sender": {"name": sender.sender_name, "email": sender.sender_email},
                "receiver": {"name": receiver.name, "email": receiver.email, "phoneNumber": receiver.phone_number},
                "deliveryPoint": {"boxMachineName": delivery_point.name},
            },
            autorefresh=True,
        )

        if resp.status == 200:
            return await resp.json()

        raise UnidentifiedAPIError(reason=resp)

    async def create_blik_session(
        self, amount: float | str, shipment_number: str, currency: str = "PLN"
    ) -> None | dict:
        """Creates blik session for sending parcel

        :param amount: amount of money that has to be paid
        :type amount: float | str
        :param shipment_number: shipment number of parcel that is being sent
        :type shipment_number: str
        :param currency: currency of `amount`
        :type currency: str
        :return: True if user has no active blik sessions else False
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        self._log.info(f"creating blik session for {shipment_number}")

        resp = await self.request(
            method="post",
            action="create blik session",
            url=create_blik_url,
            auth=True,
            headers=None,
            data={
                "shipmentNumber": shipment_number,
                "amount": amount,
                "currency": currency,
                "process": "C2X",
                "paymentMethod": "CODE",
            },
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"created blik session for {shipment_number}")
            return await resp.json()

        raise UnidentifiedAPIError(reason=resp)

    async def validate_send(
        self,
        drop_off_point: str,
        shipment_number: str | None = None,
        parcel_obj: SentParcel | None = None,
        location: dict | None = None,
    ) -> SentParcel:
        """Validates sending parcel

        :param drop_off_point: parcel machine codename where you want to drop-opp parcel
        :type drop_off_point: str
        :param shipment_number: sent parcel shipment number
        :type shipment_number: str | None
        :param parcel_obj: sent parcel object
        :type parcel_obj: SentParcel | None
        :param location: ...
        :type location: dict | None
        :return: Sent parcel with filled compartment properties
        :rtype: SentParcel
        :raises SingleParamError: Fields shipment_number and parcel_obj filled in but only one of them is required
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises NoParcelError: Could not get parcel object from provided data
        :raises MissingParamsError: None of required shipment number and parcel object params are filled
        :raises ValueError: Missing drop-off point
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        if not self.auth_token:
            self._log.error("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        parcel_obj_: SentParcel | None = None
        if shipment_number is None is parcel_obj:
            self._log.error("none of shipment_number and parcel_obj filled in")
            raise MissingParamsError(reason="None of params are filled (one required)")
        elif shipment_number is not None is not parcel_obj:
            self._log.error("shipment_number and parcel_obj filled in")
            raise SingleParamError(reason="Fields shipment_number and parcel_obj filled in! Choose one!")
        elif shipment_number:
            self._log.debug(f"parcel_obj not provided, getting from shipment number {shipment_number}")
            parcel_obj_ = await self.get_parcel(
                shipment_number=shipment_number, parcel_type=ParcelType.SENT, parse=True
            )
        elif parcel_obj:
            parcel_obj_ = parcel_obj

        if parcel_obj_ is None:
            raise NoParcelError(reason="Could not obtain parcel!")

        if parcel_obj_.drop_off_point is None:
            raise ValueError("Missing drop-off point!")

        self._log.info(f"validating send for {parcel_obj_.shipment_number}")

        resp = await self.request(
            method="post",
            action="validate send parcel data",
            url=validate_sent_url,
            auth=True,
            headers=None,
            data={
                "parcel": {
                    "shipmentNumber": parcel_obj_.shipment_number,
                    "quickSendCode": parcel_obj_.quick_send_code,
                },
                "geoPoint": location,
                "boxMachineName": drop_off_point,
            },
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"validated send for for {parcel_obj_.shipment_number}")
            parcel_obj_.compartment_properties = await resp.json()
            return parcel_obj_

        raise UnidentifiedAPIError(reason=resp)

    async def open_send_compartment(self, parcel_obj: Parcel) -> bool:
        """Opens compartment on parcel that is being sent

        :param parcel_obj: Parcel object
        :type parcel_obj: Parcel
        :return: True if successfully opened compartment else False
        :rtype: bool
        :raises NotAuthenticatedError: User not logged in (missing auth_token)
        """

        self._log.info(f"opening compartment for {parcel_obj.shipment_number}")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post",
            action=f"open send compartment for {parcel_obj.shipment_number}",
            url=open_sent_url,
            auth=True,
            headers=None,
            data={"sessionUuid": parcel_obj.compartment_properties.session_uuid},
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"opened send compartment for {parcel_obj.shipment_number}")
            return True

        return False

    async def reopen_send_compartment(self, parcel_obj: SentParcel) -> bool:
        """Reopens compartment after sending process

        :param parcel_obj: Parcel object
        :type parcel_obj: SentParcel
        :return: True if successfully reopened compartment else False
        :rtype: bool
        :raises NotAuthenticatedError: User not logged in (missing auth_token)
        """

        self._log.info(f"reopening send compartment for {parcel_obj.shipment_number}")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post",
            action=f"reopen compartment for {parcel_obj.shipment_number}",
            url=compartment_reopen_url,
            auth=True,
            headers=None,
            data={"sessionUuid": parcel_obj.compartment_properties.session_uuid},
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"reopened send compartment for {parcel_obj.shipment_number}")
            return True

        return False

    async def check_send_compartment_status(
        self, parcel_obj: SentParcel, expected_status: CompartmentExpectedStatus = CompartmentExpectedStatus.OPENED
    ) -> bool:
        """Checks and compare compartment status (e.g. opened, closed) with expected status

        :param parcel_obj: Parcel object
        :type parcel_obj: SentParcel
        :param expected_status: Compartment expected status
        :type expected_status: CompartmentExpectedStatus
        :return: True if actual status equals expected status else False
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info(f"checking compartment status for {parcel_obj.shipment_number}")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post",
            action="check send compartment status",
            url=status_sent_url,
            auth=True,
            headers=None,
            data={
                "sessionUuid": parcel_obj.compartment_properties.session_uuid,
                "expectedStatus": expected_status.name,
            },
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug(f"checked send compartment status for {parcel_obj.shipment_number}")
            parcel_obj.compartment_status = (await resp.json())["status"]
            return CompartmentExpectedStatus[(await resp.json())["status"]] == expected_status

        return False

    async def get_prices(self) -> dict:
        """Fetches prices for inpost services

        :return: :class:`dict` of prices for inpost services
        :rtype: dict
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info("getting parcel prices")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="get",
            action="get prices",
            url=parcel_prices_url,
            auth=True,
            headers=None,
            data=None,
            autorefresh=True,
        )
        if resp.status == 200:
            self._log.debug("got parcel prices")
            return await resp.json()

        raise UnidentifiedAPIError(reason=resp)

    async def get_friends(self, parse=False) -> dict | List[Friend]:
        """Fetches user friends for inpost services

        :param parse: switch for parsing response
        :type parse: bool
        :return: :class:`dict` of user friends for inpost services
        :rtype: dict
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info("getting friends")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="get", action="get friends", url=friendship_url, auth=True, headers=None, data=None, autorefresh=True
        )
        if resp.status == 200:
            self._log.debug("got user friends")
            _friends = await resp.json()
            return (
                _friends
                if not parse
                else [Friend(friend_data=friend, logger=self._log) for friend in _friends["friends"]]
            )

        raise UnidentifiedAPIError(reason=resp)

    async def get_parcel_friends(self, shipment_number: int | str, parse=False) -> dict:
        """Fetches parcel friends

        :param shipment_number: shipment number of parcel that friends are fetched
        :type shipment_number: int | str
        :param parse: switch for parsing response
        :type parse: bool
        :return: dict containing friends data
        :rtype: dict
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info("getting parcel friends")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="get",
            action="get parcel friends",
            url=f"{friendship_url}{shipment_number}",
            auth=True,
            headers=None,
            data=None,
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug("got parcel friends")
            r = await resp.json()
            if "sharedWith" in r:
                return (
                    r
                    if not parse
                    else {
                        "friends": [Friend(friend_data=friend, logger=self._log) for friend in r["friends"]],
                        "shared_with": [Friend(friend_data=friend, logger=self._log) for friend in r["sharedWith"]],
                    }
                )
            return (
                {"friends": [Friend(friend_data=friend, logger=self._log) for friend in r["friends"]]} if parse else r
            )

        raise UnidentifiedAPIError(reason=resp)

    async def add_friend(self, name: str, phone_number: str | int, code: str | int, parse=False) -> dict | Friend:
        """Adds user friends for inpost services

        :param name: name of further inpost friend
        :type name: str
        :param phone_number: further friend phone number
        :type phone_number: str | int
        :param code: used when you have a friendship code from your friend
        :type code: str | int
        :param parse: switch for parsing response
        :type parse: bool
        :return: :class:`dict` with friends details
        :rtype: dict
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: User with specified phone number not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        :raises ValueError: Name length exceeds 20 characters
        """

        self._log.info("adding user friend")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        if len(name) > 20:
            raise ValueError(f"Name too long: {name} (over 20 characters")

        if code:
            if isinstance(code, int):
                code = str(code)

            resp = await self.request(
                method="post",
                action="add friend",
                url=validate_friendship_url,
                auth=True,
                headers=None,
                data={"invitationCode": code},
                autorefresh=True,
            )

            if resp.status == 200:
                self._log.debug("added user friend")
                return await resp.json() if not parse else Friend(await resp.json(), logger=self._log)

        else:
            if isinstance(phone_number, int):
                phone_number = str(phone_number)

            resp = await self.request(
                method="post",
                action="add friend",
                url=friendship_url,
                auth=True,
                headers=None,
                data={"phoneNumber": phone_number, "name": name},
                autorefresh=True,
            )

            if resp.status == 200:
                self._log.debug("added user friend")
                r = await resp.json()
                if r["status"] == "AUTO_ACCEPT":
                    return (
                        {"phoneNumber": phone_number, "name": name}
                        if not parse
                        else Friend({"phoneNumber": phone_number, "name": name}, logger=self._log)
                    )

                elif r["status"] == "RETURN_INVITATION_CODE":
                    return r if not parse else Friend.from_invitation(invitation_data=r, logger=self._log)

                else:
                    self._log.warning(r)

        raise UnidentifiedAPIError(reason=resp)

    async def remove_friend(self, uuid: str | None, name: str | None, phone_number: str | int | None) -> bool:
        """Removes user friend for inpost services with specified `uuid`/`phone_number`/`name`

        :param uuid: uuid of inpost friend to remove
        :type uuid: str | None
        :param name: name of inpost friend to remove
        :type name: str | None
        :param phone_number: phone number of inpost friend to remove
        :type phone_number: str | int | None
        :return: True if friend is removed
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Friend not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        :raises ValueError: Name length exceeds 20 characters
        :raises MissingParamsError: none of required uuid, name or phone_number params are filled
        """

        self._log.info("removing user friend")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        if uuid is None and name is None and phone_number is None:
            raise MissingParamsError(reason="None of params are filled (one required)")

        if isinstance(phone_number, int):
            phone_number = str(phone_number)

        if uuid is None:
            f = await self.get_friends()
            if not isinstance(f, dict):
                return False

            if phone_number:
                uuid = next((friend["uuid"] for friend in f["friends"] if friend["phoneNumber"] == phone_number))
            else:
                uuid = next((friend["uuid"] for friend in f["friends"] if friend["name"] == name))

        resp = await self.request(
            method="delete",
            action="remove user friend",
            url=f"{friendship_url}{uuid}",
            auth=True,
            headers=None,
            data=None,
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug("removed user friend")
            return True

        return False

    async def update_friend(self, uuid: str | None, phone_number: str | int | None, name: str) -> bool:
        """Updates user friend for inpost services with specified `name`

        :param uuid: uuid of inpost friend to update
        :type uuid: str | None
        :param name: new name of inpost friend
        :type name: str
        :param phone_number: phone number of inpost friend to update
        :type phone_number: str | int | None
        :return: True if friend is updated
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnauthorizedError: Unauthorized access to inpost services,
        :raises NotFoundError: Friend not found
        :raises UnidentifiedAPIError: Unexpected thing happened
        :raises ValueError: Name length exceeds 20 characters
        """

        self._log.info("updating user friend")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        if len(name) > 20:
            raise ValueError(f"Name too long: {name} (over 20 characters")

        if isinstance(phone_number, int):
            phone_number = str(phone_number)

        if uuid is None:
            obtained_friends = await self.get_friends()
            if not isinstance(obtained_friends, dict):
                return False

            uuid = next((friend["uuid"] for friend in obtained_friends if friend["phoneNumber"] == phone_number))

        resp = await self.request(
            method="patch",
            action="update user friend",
            url=f"{friendship_url}{uuid}",
            auth=True,
            headers=None,
            data=None,
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug("updated user friend")
            return True

        return False

    async def share_parcel(self, uuid: str, shipment_number: int | str) -> bool:
        """Shares parcel to a pre-initialized friend

        :param uuid: uuid of inpost friend to update
        :type uuid: str
        :param shipment_number: Parcel's shipment number
        :type shipment_number: int | str
        :return: True if parcel is shared
        :rtype: bool
        :raises NotAuthenticatedError: User not authenticated in inpost service
        :raises UnidentifiedAPIError: Unexpected thing happened
        """

        self._log.info(f"sharing parcel: {shipment_number}")

        if not self.auth_token:
            self._log.debug("authorization token missing")
            raise NotAuthenticatedError(reason="Not logged in")

        resp = await self.request(
            method="post",
            action=f"share parcel: {shipment_number}",
            url=shared_url,
            auth=True,
            headers=None,
            data={
                "parcels": [{"shipmentNumber": shipment_number, "friendUuids": [uuid]}],
            },
            autorefresh=True,
        )

        if resp.status == 200:
            self._log.debug("updated user friend")
            return True

        return False



from .endpoints import (
    accept_friendship_url,
    blik_status_url,
    collect_url,
    compartment_open_url,
    compartment_reopen_url,
    compartment_status_url,
    confirm_sent_url,
    confirm_sms_code_url,
    create_blik_url,
    create_url,
    friendship_url,
    login_url,
    logout_url,
    multi_url,
    open_sent_url,
    parcel_notifications_url,
    parcel_points_url,
    parcel_prices_url,
    refresh_token_url,
    reopen_sent_url,
    returns_url,
    send_sms_code_url,
    sent_url,
    shared_url,
    status_sent_url,
    terminate_collect_session_url,
    tickets_url,
    tracked_url,
    validate_friendship_url,
    validate_sent_url,
)
from .exceptions import (
    MissingParamsError,
    NoParcelError,
    NotAuthenticatedError,
    NotFoundError,
    ParcelTypeError,
    PhoneNumberError,
    ReAuthenticationError,
    RefreshTokenError,
    SingleParamError,
    SmsCodeError,
    UnauthorizedError,
    UnidentifiedAPIError,
    UnidentifiedError,
    UnidentifiedParcelError,
    UserLocationError,
)
from .friends import Friend
from .headers import appjson
from .notifications import Notification
from .parcels import (
    CompartmentLocation,
    CompartmentProperties,
    EventLog,
    MultiCompartment,
    Operations,
    Parcel,
    PickupPoint,
    Point,
    QRCode,
    Receiver,
    ReturnParcel,
    Sender,
    SentParcel,
    SharedTo,
)
from .statuses import (
    CompartmentActualStatus,
    CompartmentExpectedStatus,
    DeliveryType,
    ParcelAdditionalInsurance,
    ParcelCarrierSize,
    ParcelDeliveryType,
    ParcelLockerSize,
    ParcelOwnership,
    ParcelPointOperations,
    ParcelServiceName,
    ParcelShipmentType,
    ParcelStatus,
    ParcelType,
    PaymentStatus,
    PaymentType,
    PointType,
    ReturnsStatus,
)

__all__ = [
    "accept_friendship_url",
    "blik_status_url",
    "collect_url",
    "compartment_open_url",
    "compartment_reopen_url",
    "compartment_status_url",
    "confirm_sent_url",
    "confirm_sms_code_url",
    "create_url",
    "create_blik_url",
    "friendship_url",
    "login_url",
    "logout_url",
    "multi_url",
    "open_sent_url",
    "tracked_url",
    "parcel_notifications_url",
    "parcel_points_url",
    "parcel_prices_url",
    "parcels",
    "refresh_token_url",
    "reopen_sent_url",
    "returns_url",
    "send_sms_code_url",
    "sent_url",
    "shared_url",
    "status_sent_url",
    "terminate_collect_session_url",
    "tickets_url",
    "validate_friendship_url",
    "validate_sent_url",
    "MissingParamsError",
    "NoParcelError",
    "NotAuthenticatedError",
    "NotFoundError",
    "ParcelTypeError",
    "PhoneNumberError",
    "ReAuthenticationError",
    "RefreshTokenError",
    "SingleParamError",
    "SmsCodeError",
    "UnauthorizedError",
    "UnidentifiedAPIError",
    "UnidentifiedError",
    "UnidentifiedParcelError",
    "UserLocationError",
    "Friend",
    "appjson",
    "Notification",
    "CompartmentLocation",
    "CompartmentProperties",
    "EventLog",
    "MultiCompartment",
    "Parcel",
    "PickupPoint",
    "Point",
    "QRCode",
    "Receiver",
    "ReturnParcel",
    "Sender",
    "SentParcel",
    "SharedTo",
    "CompartmentActualStatus",
    "CompartmentExpectedStatus",
    "DeliveryType",
    "ParcelAdditionalInsurance",
    "ParcelCarrierSize",
    "ParcelDeliveryType",
    "ParcelLockerSize",
    "ParcelOwnership",
    "ParcelPointOperations",
    "ParcelServiceName",
    "ParcelShipmentType",
    "ParcelStatus",
    "ParcelType",
    "PaymentType",
    "PointType",
    "ReturnsStatus",
]



from enum import Enum, EnumMeta


class Meta(EnumMeta):  # temporary handler for unexpected keys in enums
    def __getitem__(cls, item):
        try:
            return super().__getitem__(item) if item is not None else None
        except KeyError:
            return cls.UNKNOWN

    def __getattribute__(cls, item):
        try:
            return super().__getattribute__(item) if item is not None else None
        except KeyError:
            return cls.UNKNOWN

    # def get_all(cls):
    #     return [getattr(cls, name) for name in cls.__members__]
    #
    # def get_without(cls, without: "ParcelBase" | List["ParcelBase"]):
    #     if isinstance(without, ParcelBase):
    #         without = [without]
    #
    #     return [element for element in cls.get_all() if element not in without]


class ParcelBase(Enum, metaclass=Meta):
    """Base :class:`Enum` class to derive from"""

    def __gt__(self, other):
        if isinstance(other, ParcelBase):
            ...

        return False

    def __ge__(self, other):
        if isinstance(other, ParcelBase):
            ...

        return False

    def __le__(self, other):
        if isinstance(other, ParcelBase):
            ...

        return False

    def __lt__(self, other):
        if isinstance(other, ParcelBase):
            ...

        return False

    def __eq__(self, other):
        if isinstance(other, ParcelBase):
            return self.name == other.name

        return False

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items())
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class ParcelCarrierSize(ParcelBase):
    """:class:`Enum` that holds parcel size for carrier shipment type"""

    UNKNOWN = "UNKNOWN DATA"
    A = "8x38x64"
    B = "19x38x64"
    C = "41x38x64"
    D = "50x50x80"
    OTHER = "UNKNOWN DIMENSIONS"


class ParcelLockerSize(ParcelBase):
    """:class:`Enum` that holds parcel size for parcel locker shipment type"""

    UNKNOWN = "UNKNOWN DATA"
    A = "8x38x64"
    B = "19x38x64"
    C = "41x38x64"


class ParcelDeliveryType(ParcelBase):
    """:class:`Enum` that holds parcel delivery types"""

    UNKNOWN = "UNKNOWN DATA"
    parcel_locker = "Paczkomat"
    courier = "Kurier"
    parcel_point = "PaczkoPunkt"


class ParcelShipmentType(ParcelBase):
    """:class:`Enum` that holds parcel shipment types"""

    UNKNOWN = "UNKNOWN DATA"
    parcel = "Paczkomat"
    courier = "Kurier"
    parcel_point = "PaczkoPunkt"


class ParcelAdditionalInsurance(ParcelBase):
    UNKNOWN = "UNKNOWN DATA"
    UNINSURANCED = 1
    ONE = 2  # UPTO 5000
    TWO = 3  # UPTO 10000
    THREE = 4  # UPTO 20000


class ParcelType(ParcelBase):
    """:class:`Enum` that holds parcel types"""

    UNKNOWN = "UNKNOWN DATA"
    TRACKED = "Przychodzące"
    SENT = "Wysłane"
    RETURNS = "Zwroty"


class PointType(ParcelBase):
    """:class: `Enum` that holds point types"""

    # TODO: get known what does superpop stand for
    UNKNOWN = "UNNKOWN DATA"
    PL = "Paczkomat"
    parcel_locker_superpop = "some paczkomat or pok stuff"
    POK = "Mobilny punkt obsługi klienta"
    POP = "Punkt odbioru paczki"


class ParcelPointOperations(ParcelBase):
    """:class: `Enum` that holds parcel operation types"""

    # TODO: Probably missing something, recheck needed
    UNKNOWN = "UNNKOWN DATA"
    CREATE = "c2x-target"
    SEND = "remote-send"


class ParcelStatus(ParcelBase):
    """:class:`Enum` that holds parcel statuses"""

    UNKNOWN = "UNKNOWN DATA"
    CREATED = "W trakcie przygotowania"  # TODO: translate from app
    OFFERS_PREPARED = "Oferty przygotowane"  # TODO: translate from app
    OFFER_SELECTED = "Oferta wybrana"  # TODO: translate from app
    CONFIRMED = "Potwierdzona"
    READY_TO_PICKUP_FROM_POK = "Gotowa do odbioru w PaczkoPunkcie"
    OVERSIZED = "Gabaryt"
    DISPATCHED_BY_SENDER_TO_POK = "Nadana w PaczkoPunkcie"
    DISPATCHED_BY_SENDER = "Nadana w paczkomacie"
    COLLECTED_FROM_SENDER = "Odebrana od nadawcy"
    TAKEN_BY_COURIER = "Odebrana przez Kuriera"
    ADOPTED_AT_SOURCE_BRANCH = "Przyjęta w oddziale"
    SENT_FROM_SOURCE_BRANCH = "Wysłana z oddziału"
    READDRESSED = "Zmiana punktu dostawy"  # TODO: translate from app
    OUT_FOR_DELIVERY = "Wydana do doręczenia"
    READY_TO_PICKUP = "Gotowa do odbioru"
    PICKUP_REMINDER_SENT = "Wysłano przypomnienie o odbiorze"  # TODO: translate from app
    PICKUP_TIME_EXPIRED = "Upłynął czas odbioru"  # TODO: translate from app
    AVIZO = "Powrót do oddziału"
    TAKEN_BY_COURIER_FROM_POK = "Odebrana z PaczkoPunktu nadawczego"
    REJECTED_BY_RECEIVER = "Odrzucona przez odbiorcę"  # TODO: translate from app
    UNDELIVERED = "Nie dostarczona"  # TODO: translate from app
    DELAY_IN_DELIVERY = "Opóźnienie w dostarczeniu"  # TODO: translate from app
    RETURNED_TO_SENDER = "Zwrócona do nadawcy"  # TODO: translate from app
    READY_TO_PICKUP_FROM_BRANCH = "Gotowa do odbioru z oddziału"  # TODO: translate from app
    DELIVERED = "Doręczona"
    CANCELED = "Anulowana"  # TODO: translate from app
    CLAIMED = "Zareklamowana"
    STACK_IN_CUSTOMER_SERVICE_POINT = "Przesyłka magazynowana w punkcie obsługi klienta"  # TODO: translate from app
    STACK_PARCEL_PICKUP_TIME_EXPIRED = "Upłynął czas odbioru"  # TODO: translate from app
    UNSTACK_FROM_CUSTOMER_SERVICE_POINT = "?"  # TODO: translate from app
    COURIER_AVIZO_IN_CUSTOMER_SERVICE_POINT = "Przekazana do punktu obsługi klienta"  # TODO: translate from app
    TAKEN_BY_COURIER_FROM_CUSTOMER_SERVICE_POINT = (
        "Odebrana przez kuriera z punktu obsługi klienta"  # TODO: translate from app
    )
    STACK_IN_BOX_MACHINE = "Przesyłka magazynowana w paczkomacie tymczasowym"
    STACK_PARCEL_IN_BOX_MACHINE_PICKUP_TIME_EXPIRED = "Upłynął czas odbioru z paczkomatu"  # TODO: translate from app
    UNSTACK_FROM_BOX_MACHINE = "Odebrana z paczkomatu"  # TODO: translate from app
    ADOPTED_AT_SORTING_CENTER = "Przyjęta w sortowni"
    OUT_FOR_DELIVERY_TO_ADDRESS = "Gotowa do doręczenia"
    PICKUP_REMINDER_SENT_ADDRESS = "Wysłano przypomnienie o odbiorze"  # TODO: translate from app
    UNDELIVERED_WRONG_ADDRESS = "Nie dostarczono z powodu złego adresu"  # TODO: translate from app
    UNDELIVERED_COD_CASH_RECEIVER = "Nie dostarczono z powodu nieopłacenia"  # TODO: translate from app
    REDIRECT_TO_BOX = "Przekierowana do paczkomatu"  # TODO: translate from app
    CANCELED_REDIRECT_TO_BOX = "Anulowano przekierowanie do paczkomatu"  # TODO: translate from app


class DeliveryType(ParcelBase):
    # TODO: look for more types
    UNKNOWN = "UNKNOWN DATA"
    BOX_MACHINE = "Paczkomat"


class ReturnsStatus(ParcelBase):
    # TODO: translate from app and fill missing ones
    ACCEPTED = "Zaakceptowano"
    USED = "Nadano"
    DELIVERED = "Dostarczono"
    UNKNOWN = "UNKNOWN DATA"


class ParcelOwnership(ParcelBase):
    """:class:`Enum` that holds parcel ownership types"""

    UNKNOWN = "UNKNOWN DATA"
    FRIEND = "Zaprzyjaźniona"
    OWN = "Własna"


# both are the same, only for being clear
class CompartmentExpectedStatus(ParcelBase):
    """:class:`Enum` that holds compartment expected statuses"""

    UNKNOWN = "UNKNOWN DATA"
    OPENED = "Otwarta"
    CLOSED = "Zamknięta"


class CompartmentActualStatus(ParcelBase):
    """:class:`Enum` that holds compartment actual statuses"""

    UNKNOWN = "UNKNOWN DATA"
    OPENED = "Otwarta"
    CLOSED = "Zamknięta"


class PaymentType(ParcelBase):
    UNKNOWN = "UNKNOWN DATA"
    NOTSUPPORTED = "Payments are not supported"  # klucz 0
    BY_CARD_IN_MACHINE = "Payment by card in the machine"  # klucz 2


class PaymentStatus(ParcelBase):
    UNKNOWN = "UNKNOWN DATA"
    C2X_COMPLETED = "Completed"


class ParcelServiceName(ParcelBase):
    UNKNOWN = "UNKNOWN DATA"
    ALLEGRO_PARCEL = 1
    ALLEGRO_PARCEL_SMART = 2
    ALLEGRO_LETTER = 3
    ALLEGRO_COURIER = 4
    STANDARD = 5
    STANDARD_PARCEL_SMART = 6
    PASS_THRU = 7
    CUSTOMER_SERVICE_POINT = 8
    REVERSE = 9
    STANDARD_COURIER = 10
    REVERSE_RETURN = 11



import logging
import random
from io import BytesIO
from typing import List, Tuple

import qrcode
from arrow import arrow, get

from inpost.static.exceptions import UnknownStatusError
from inpost.static.statuses import (
    CompartmentActualStatus,
    ParcelCarrierSize,
    ParcelDeliveryType,
    ParcelLockerSize,
    ParcelOwnership,
    ParcelShipmentType,
    ParcelStatus,
    PaymentStatus,
    PaymentType,
    PointType,
    ReturnsStatus,
)


class BaseParcel:
    """Object representation of :class:`inpost.api.Inpost` parcel base. Gather things shared by all parcel types.

    :param parcel_data: :class:`dict` containing all parcel data
    :type parcel_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, parcel_data: dict, logger: logging.Logger):
        self.shipment_number = parcel_data.get("shipmentNumber")
        self._log: logging.Logger = logger.getChild(f"{self.__class__.__name__}.{self.shipment_number}")
        self.status: ParcelStatus = ParcelStatus[parcel_data.get("status")]
        self.expiry_date: arrow | None = get(parcel_data["expiryDate"]) if "expiryDate" in parcel_data else None
        self.operations: Operations = Operations(operations_data=parcel_data["operations"], logger=self._log)
        self.event_log: List[EventLog] = [
            EventLog(eventlog_data=event, logger=self._log) for event in parcel_data["eventLog"]
        ]


class Parcel(BaseParcel):
    """Object representation of :class:`inpost.api.Inpost` incoming parcel

    :param parcel_data: :class:`dict` containing all parcel data
    :type parcel_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, parcel_data: dict, logger: logging.Logger):
        """Constructor method

        :param parcel_data: dict containing parcel data
        :type parcel_data: dict
        :param logger: logger instance
        :type logger: logging.Logger
        """

        super().__init__(parcel_data, logger)
        self._log: logging.Logger = logger.getChild(f"{self.__class__.__name__}.{self.shipment_number}")
        self.shipment_type: ParcelShipmentType = ParcelShipmentType[parcel_data.get("shipmentType")]
        self._open_code: str | None = parcel_data.get("openCode", None)
        self._qr_code: QRCode | None = (
            QRCode(qrcode_data=parcel_data["qrCode"], logger=self._log) if "qrCode" in parcel_data else None
        )
        self.stored_date: arrow | None = get(parcel_data["storedDate"]) if "storedDate" in parcel_data else None
        self.pickup_date: arrow | None = get(parcel_data["pickUpDate"]) if "pickUpDate" in parcel_data else None
        self.parcel_size: ParcelLockerSize | ParcelCarrierSize = (
            ParcelLockerSize[parcel_data.get("parcelSize")]
            if self.shipment_type == ParcelShipmentType.parcel
            else ParcelCarrierSize[parcel_data.get("parcelSize")]
        )
        self.receiver: Receiver | None = (
            Receiver(receiver_data=parcel_data["receiver"], logger=self._log) if "receiver" in parcel_data else None
        )
        self.sender: Sender | None = (
            Sender(sender_data=parcel_data["sender"], logger=self._log) if "sender" in parcel_data else None
        )
        self.pickup_point: PickupPoint | None = (
            PickupPoint(point_data=parcel_data["pickUpPoint"], logger=self._log)
            if "pickUpPoint" in parcel_data
            else None
        )
        self.multi_compartment: MultiCompartment | None = (
            MultiCompartment(parcel_data["multiCompartment"], logger=self._log)
            if "multiCompartment" in parcel_data
            else None
        )

        self.is_end_off_week_collection: bool | None = parcel_data.get("endOfWeekCollection", None)
        self.status: ParcelStatus | None = ParcelStatus[parcel_data.get("status")]
        self.avizo_transaction_status: str | None = parcel_data.get("avizoTransactionStatus", None)
        self.shared_to: List[SharedTo] | None = (
            [SharedTo(sharedto_data=person, logger=self._log) for person in parcel_data["sharedTo"]]
            if "sharedTo" in parcel_data
            else None
        )
        self.ownership_status: ParcelOwnership | None = ParcelOwnership[parcel_data.get("ownershipStatus")]
        self.economy_parcel: bool | None = parcel_data.get("economyParcel", None)
        self._compartment_properties: CompartmentProperties | None = None

        self._log.debug(f"created parcel with shipment number {self.shipment_number}")

        # log all unexpected things, so you can make an issue @github
        if self.shipment_type == ParcelShipmentType.UNKNOWN:
            self._log.warning(f'unexpected shipment_type: {parcel_data["shipmentType"]}')

        if self.parcel_size == ParcelCarrierSize.UNKNOWN or self.parcel_size == ParcelLockerSize.UNKNOWN:
            self._log.warning(f'unexpected parcel_size: {parcel_data["parcelSize"]}')

        if self.status == ParcelStatus.UNKNOWN:
            self._log.warning(f'unexpected parcel status: {parcel_data["status"]}')

        if self.ownership_status == ParcelOwnership.UNKNOWN:
            self._log.warning(f'unexpected ownership status: {parcel_data["ownershipStatus"]}')

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")

    def __str__(self):
        return (
            f"Sender: {str(self.sender)}\n"
            f"Shipment number: {self.shipment_number}\n"
            f"Status: {self.status}\n"
            f"Pickup point: {self.pickup_point}"
        )

    @property
    def open_code(self) -> str | None:
        """Returns an open code for :class:`Parcel`

        :return: Open code for :class:`Parcel`
        :rtype: str
        """

        self._log.debug("getting open code")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got open code")
            return self._open_code

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @property
    def generate_qr_image(self) -> BytesIO | None:
        """Returns a QR image for :class:`Parcel`

        :return: QR image for :class:`Parcel`
        :rtype: BytesIO
        """

        self._log.debug("generating qr image")
        if self.shipment_type == ParcelShipmentType.parcel and self._qr_code is not None:
            self._log.debug("got qr image")
            return self._qr_code.qr_image

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @property
    def compartment_properties(self):
        """Returns a compartment properties for :class:`Parcel`

        :return: Compartment properties for :class:`Parcel`
        :rtype: CompartmentProperties
        """

        self._log.debug("getting comparment properties")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got compartment properties")
            return self._compartment_properties

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @compartment_properties.setter
    def compartment_properties(self, compartmentproperties_data: dict):
        """Set compartment properties for :class:`Parcel`

        :param compartmentproperties_data: :class:`dict` containing compartment properties data for :class:`Parcel`
        :type compartmentproperties_data: dict
        """

        self._log.debug(f"setting compartment properties with {compartmentproperties_data}")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("compartment properties set")
            self._compartment_properties = CompartmentProperties(
                compartmentproperties_data=compartmentproperties_data, logger=self._log
            )
            return

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")

    @property
    def compartment_location(self):
        """Returns a compartment location for :class:`Parcel`

        :return: Compartment location for :class:`Parcel`
        :rtype: CompartmentLocation
        """

        self._log.debug("getting compartment location")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got compartment location")
            return self._compartment_properties.location if self._compartment_properties else None

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @compartment_location.setter
    def compartment_location(self, location_data: dict):
        """Set compartment location for :class:`Parcel`

        :param location_data: :class:`dict` containing `compartment properties` data for :class:`Parcel`
        :type location_data: dict
        """

        self._log.debug(f"setting compartment location with {location_data}")
        if self.shipment_type == ParcelShipmentType.parcel and self._compartment_properties is not None:
            self._log.debug("compartment location set")
            self._compartment_properties.location = location_data
            return

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")

    @property
    def compartment_status(self) -> CompartmentActualStatus | None:
        """Returns a compartment status for :class:`Parcel`

        :return: Compartment status for :class:`Parcel`
        :rtype: CompartmentActualStatus
        """

        self._log.debug("getting compartment status")

        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got compartment status")
            return self._compartment_properties.status if self._compartment_properties else None

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @compartment_status.setter
    def compartment_status(self, status) -> None:
        """Set compartment location for :class:`SentParcel`

        :param status: compartment properties status for :class:`SentParcel`
        :type status: str | CompartmentActualStatus
        """

        self._log.debug(f"setting compartment status with {status}")
        if self._compartment_properties is None:
            self._log.warning("tried to assign status to empty _compartment_properties")
            return

        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("compartment status set")
            self._compartment_properties.status = status
            return

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")

    @property
    def compartment_open_data(self) -> dict | None:
        """Returns a compartment open data for :class:`Parcel`

        :return: dict containing compartment open data for :class:`Parcel`
        :rtype: dict
        """

        self._log.debug("getting compartment open data")
        if self.receiver is None:
            return None

        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got compartment open data")
            return {
                "shipmentNumber": self.shipment_number,
                "openCode": self._open_code,
                "receiverPhoneNumber": self.receiver.phone_number,
            }

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @property
    def mocked_location(self) -> dict | None:
        """Returns a mocked location for :class:`Parcel`

        :return: dict containing mocked location for :class:`Parcel or None if wrong parcel shipment type`
        :rtype: dict | None
        """

        self._log.debug("getting mocked location")
        if self.pickup_point is None:
            return None
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got mocked location")
            return {
                "latitude": round(self.pickup_point.latitude + random.uniform(-0.00005, 0.00005), 6),
                "longitude": round(self.pickup_point.longitude + random.uniform(-0.00005, 0.00005), 6),
                "accuracy": round(random.uniform(1, 4), 1),
            }

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @property
    def is_multicompartment(self) -> bool:
        """Specifies if parcel is in multi compartment

        :return: True if parcel is in multicompartment
        :rtype: bool
        """

        return self.multi_compartment is not None

    @property
    def is_main_multicompartment(self) -> bool | None:
        """Specifies if parcel is main parcel in multi compartment

        :return: True if parcel is in multicompartment
        :rtype: bool
        """

        if self.is_multicompartment and self.multi_compartment is not None:
            return self.multi_compartment.shipment_numbers is not None

        return None

    @property
    def has_airsensor(self) -> bool | None:
        if self.pickup_point is not None:
            return self.pickup_point.air_sensor_data is not None

        return None


class ReturnParcel(BaseParcel):
    # TODO: Prepare properties required to ease up access
    """Object representation of :class:`inpost.api.Inpost` returned parcel

    :param parcel_data: :class:`dict` containing all parcel data
    :type parcel_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, parcel_data: dict, logger: logging.Logger):
        """Constructor method

        :param parcel_data: dict containing parcel data
        :type parcel_data: dict
        :param logger: logger instance
        :type logger: logging.Logger
        """

        super().__init__(parcel_data, logger)
        self.uuid: str = parcel_data["uuid"]
        self.rma: str = parcel_data["rma"]
        self.organization_name: str = parcel_data["organizationName"]
        self.created_date: arrow = get(parcel_data["createdDate"])
        self.accepted_date: arrow = get(parcel_data["acceptedDate"])
        self.expiry_date: arrow = get(parcel_data["expiryDate"])
        self.sent_date: arrow = get(parcel_data["sentDate"])
        self.delivered_date: arrow = get(parcel_data["deliveredDate"])
        self.order_number: str = parcel_data["orderNumber"]
        self.form_type: str = parcel_data["formType"]


class SentParcel(BaseParcel):
    # TODO: Recheck properties
    """Object representation of :class:`inpost.api.Inpost` sent parcel

    :param parcel_data: :class:`dict` containing all parcel data
    :type parcel_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, parcel_data: dict, logger: logging.Logger):
        """Constructor method

        :param parcel_data: dict containing parcel data
        :type parcel_data: dict
        :param logger: logger instance
        :type logger: logging.Logger
        """

        super().__init__(parcel_data, logger)
        self.origin_system: str = parcel_data.get("originSystem", None)
        self.quick_send_code: int = parcel_data.get("quickSendCode", None)
        self._qr_code: QRCode | None = (
            QRCode(qrcode_data=parcel_data["qrCode"], logger=self._log) if "qrCode" in parcel_data else None
        )
        self.confirmation_date: arrow = get(parcel_data.get("confirmationDate", None))
        self.shipment_type: ParcelShipmentType = ParcelShipmentType[parcel_data["shipmentType"]]
        self.parcel_size: ParcelLockerSize | ParcelCarrierSize = (
            ParcelLockerSize[parcel_data.get("parcelSize")]
            if self.shipment_type == ParcelShipmentType.parcel
            else ParcelCarrierSize[parcel_data.get("parcelSize")]
        )
        self.receiver: Receiver | None = (
            Receiver(receiver_data=parcel_data["receiver"], logger=self._log) if "receiver" in parcel_data else None
        )
        self.sender: Sender | None = (
            Sender(sender_data=parcel_data["sender"], logger=self._log) if "sender" in parcel_data else None
        )
        self.pickup_point: PickupPoint | None = (
            PickupPoint(point_data=parcel_data["pickUpPoint"], logger=self._log)
            if "pickUpPoint" in parcel_data
            else None
        )
        self.delivery_point: DeliveryPoint | None = (
            DeliveryPoint(parcel_data["deliveryPoint"], logger=self._log) if "deliveryPoint" in parcel_data else None
        )
        self.drop_off_point: DropOffPoint | None = (
            DropOffPoint(point_data=parcel_data["dropOffPoint"], logger=self._log)
            if "dropOffPoint" in parcel_data
            else None
        )
        self.payment: Payment | None = (
            Payment(payment_details=parcel_data["payment"], logger=self._log) if "payment" in parcel_data else None
        )
        self.unlabeled: bool = parcel_data.get("unlabeled", None)
        self.is_end_off_week_collection: bool | None = parcel_data.get("endOfWeekCollection", None)
        self.status: ParcelStatus | None = ParcelStatus[parcel_data.get("status")]
        self._compartment_properties: CompartmentProperties | None = None

    @property
    def compartment_properties(self):
        """Returns a compartment properties for :class:`SentParcel`

        :return: Compartment properties for :class:`SentParcel`
        :rtype: CompartmentProperties
        """

        self._log.debug("getting compartment properties")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got compartment properties")
            return self._compartment_properties

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @compartment_properties.setter
    def compartment_properties(self, compartmentproperties_data: dict):
        """Set compartment properties for :class:`SentParcel`

        :param compartmentproperties_data: :class:`dict` containing compartment properties data for :class:`SentParcel`
        :type compartmentproperties_data: dict
        """

        self._log.debug(f"setting compartment properties with {compartmentproperties_data}")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("compartment properties set")
            self._compartment_properties = CompartmentProperties(
                compartmentproperties_data=compartmentproperties_data, logger=self._log
            )
            return

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")

    @property
    def compartment_location(self):
        """Returns a compartment location for :class:`SentParcel`

        :return: Compartment location for :class:`SentParcel`
        :rtype: CompartmentLocation
        """

        self._log.debug("getting compartment location")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got compartment location")
            return self._compartment_properties.location if self._compartment_properties else None

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @compartment_location.setter
    def compartment_location(self, location_data: dict):
        """Set compartment location for :class:`SentParcel`
        :param location_data: :class:`dict` containing `compartment properties` data for :class:`Parcel`
        :type location_data: dict
        """

        self._log.debug(f"setting compartment location with {location_data}")
        if self._compartment_properties is None:
            self._log.warning("tried to assign location to empty _compartment_properties")
            return

        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("compartment location set")
            self._compartment_properties.location = location_data
            return

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")

    @property
    def compartment_status(self) -> CompartmentActualStatus | None:
        """Returns a compartment status for :class:`SentParcel`

        :return: Compartment status for :class:`SentParcel`
        :rtype: CompartmentActualStatus
        """

        self._log.debug("getting compartment status")
        if self._compartment_properties is None:
            self._log.warning("tried to access empty _compartment_properties")
            return None

        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got compartment status")
            return self._compartment_properties.status if self._compartment_properties else None

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None

    @compartment_status.setter
    def compartment_status(self, status: str | CompartmentActualStatus):
        """Set compartment location for :class:`SentParcel`

        :param status: compartment properties status for :class:`SentParcel`
        :type status: str | CompartmentActualStatus
        """
        if self._compartment_properties is None:
            self._log.warning("tried to assign status to empty _compartment_properties")
            return

        self._log.debug(f"setting compartment status with {status}")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("compartment status set")
            self._compartment_properties.status = status
            return

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")

    @property
    def mocked_location(self):
        """Returns a mocked location for :class:`SentParcel`

        :return: dict containing mocked location for :class:`SentParcel`
        :rtype: dict
        """

        self._log.debug("getting mocked location")
        if self.shipment_type == ParcelShipmentType.parcel:
            self._log.debug("got mocked location")
            return {
                "latitude": round(self.pickup_point.latitude + random.uniform(-0.00005, 0.00005), 6),
                "longitude": round(self.pickup_point.longitude + random.uniform(-0.00005, 0.00005), 6),
                "accuracy": round(random.uniform(1, 4), 1),
            }

        self._log.warning(f"wrong ParcelShipmentType: {repr(self.shipment_type)}")
        return None


class Receiver:
    """Object representation of :class:`BaseParcel` receiver

    :param receiver_data: :class:`dict` containing sender data for :class:`BaseParcel` and it's subclasses
    :type receiver_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, receiver_data: dict, logger: logging.Logger):
        """Constructor method

        :param receiver_data: dict containing receiver data
        :type receiver_data: dict
        :param logger: logger instance
        :type logger: logging.Logger
        """

        self.email: str = receiver_data["email"]
        self.phone_number: str = receiver_data["phoneNumber"]
        self.name: str = receiver_data["name"]
        self._log: logging.Logger = logger.getChild(self.__class__.__name__)

        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class Sender:
    """Object representation of :class:`BaseParcel` sender and it's subclasses

    :param sender_data: :class:`dict` containing sender data for :class:`BaseParcel` and it's subclasses
    :type sender_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, sender_data: dict, logger: logging.Logger):
        """Constructor method

        :param sender_data: dict containing sender data
        :type sender_data: dict
        :param logger: logger instance
        :type logger: logging.Logger
        """

        self.sender_name: str = sender_data.get("name", None)
        self.sender_email: str = sender_data.get("email", None)
        self._log: logging.Logger = logger.getChild(self.__class__.__name__)

        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")

    def __str__(self) -> str:
        return self.sender_name


class Point:
    """Object representation of :class:`BaseParcel` point and it's subclasses

    :param point_data: :class:`dict` containing point data for :class:`BaseParcel` and it's subclasses
    :type point_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, point_data: dict, logger: logging.Logger):
        """Constructor method

        :param point_data: :class:`dict` containing point data for :class:`BaseParcel` and it's subclasses
        :type point_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self.name: str = point_data["name"]
        self.latitude: float = point_data["location"]["latitude"]
        self.longitude: float = point_data["location"]["longitude"]
        self.description: str = point_data["locationDescription"]
        self.opening_hours: str = point_data["openingHours"]
        self.post_code: str = point_data["addressDetails"]["postCode"]
        self.city: str = point_data["addressDetails"]["city"]
        self.province: str = point_data["addressDetails"]["province"]
        self.street: str = point_data["addressDetails"]["street"]
        self.building_number: str = point_data["addressDetails"]["buildingNumber"]
        self.payment_type: List[PaymentType] | None = (
            [PaymentType[pt] for pt in point_data["paymentType"]] if "paymentType" in point_data else None
        )
        self.virtual: int = point_data["virtual"]
        self.point_type: PointType | None = PointType[point_data.get("pointType")]
        self.type: List[ParcelDeliveryType] = (
            [ParcelDeliveryType[data] for data in point_data["type"]] if "type" in point_data else None
        )
        self.location_round_the_clock: bool = point_data["location247"]
        self.doubled: bool = point_data["doubled"]
        self.image_url: str = point_data["imageUrl"]
        self.easy_access_zone: bool = point_data["easyAccessZone"]
        self.air_sensor: bool = point_data["airSensor"]
        self.air_sensor_data: AirSensorData | None = (
            AirSensorData(point_data["airSensorData"], self._log) if "airSensorData" in point_data else None
        )
        self.remote_send: bool = point_data.get("remoteSend", None)
        self.remote_return: bool = point_data.get("remoteReturn", None)

        self._log.debug("created")
        if ParcelDeliveryType.UNKNOWN in self.type:
            self._log.warning(f'unknown delivery type: {point_data["type"]}')

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")

    def __str__(self) -> str:
        return self.name

    @property
    def location(self) -> Tuple[float, float]:
        """Returns a mocked location for :class:`Point`

        :return: tuple containing location for :class:`Point`
        :rtype: tuple
        """

        self._log.debug("getting location")
        return self.latitude, self.longitude


class PickupPoint(Point):
    """Object representation of :class:`BaseParcel` and it's subclasses pick up point

    :param point_data: :class:`dict` containing pickup point data for :class:`BaseParcel` and it's subclasses
    :type point_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, point_data: dict, logger: logging.Logger):
        """Constructor method

        :param point_data: :class:`dict` containing pickup point data for :class:`BaseParcel` and it's subclasses
        :type point_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """
        super().__init__(point_data, logger)


class DropOffPoint(Point):
    """Object representation of :class:`BaseParcel` and it's subclasses drop off point

    :param point_data: :class:`dict` containing pickup point data for :class:`BaseParcel` and it's subclasses
    :type point_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, point_data: dict, logger: logging.Logger):
        """Constructor method

        :param point_data: :class:`dict` containing pickup point data for :class:`BaseParcel` and it's subclasses
        :type point_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        super().__init__(point_data, logger)


class DeliveryPoint:
    """Object representation of :class:`BaseParcel` and it's subclasses delivery point

    :param delivery_point: :class:`dict` containing delivery point data for :class:`BaseParcel` and it's subclasses delivery point
    :type delivery_point: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, delivery_point: dict, logger: logging.Logger):
        """Constructor method

        :param delivery_point: :class:`dict` containing delivery point data for :class:`BaseParcel`
        and it's subclasses delivery point
        :type delivery_point: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.name = delivery_point.get("name")
        self.company_name = delivery_point.get("companyName")
        self.post_code = delivery_point["address"]["postCode"] if "address" in delivery_point else None
        self.city = delivery_point["address"]["city"] if "address" in delivery_point else None
        self.street = delivery_point["address"]["street"] if "address" in delivery_point else None
        self.building_number = delivery_point["address"]["buildingNumber"] if "address" in delivery_point else None
        self.flat_numer = delivery_point["address"]["flatNumber"] if "address" in delivery_point else None

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class Payment:
    """Object representation of :class:`BaseParcel` and it's subclasses payment

    :param payment_details: :class:`dict` containing payment data for :class:`BaseParcel` and it's subclasses payment
    :type payment_details: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, payment_details: dict, logger: logging.Logger):
        """Constructor method

        :param payment_details: :class:`dict` containing payment data for :class:`BaseParcel` and it's subclasses payment
        :type payment_details: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.paid = payment_details.get("paid")
        self.total_price = payment_details.get("totalPrice")
        self.insurance_price = payment_details.get("insurancePrice")
        self.end_of_week_collection_price = payment_details.get("endOfWeekCollectionPrice")
        self.shipment_discounted = payment_details.get("shipmentDiscounted")
        self.transaction_status = payment_details.get("transactionStatus")

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class MultiCompartment:
    """Object representation of :class:`Parcel` `multicompartment`

    :param multicompartment_data: :class:`dict` containing multicompartment data for :class:`Parcel`
    :type multicompartment_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, multicompartment_data: dict, logger: logging.Logger):
        """Constructor method:param multicompartment_data: :class:`dict` containing multicompartment data for :class:`Parcel`

        :type multicompartment_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.uuid = multicompartment_data["uuid"]
        self.shipment_numbers: List[str] | None = (
            multicompartment_data["shipmentNumbers"] if "shipmentNumbers" in multicompartment_data else None
        )
        self.presentation: bool = multicompartment_data["presentation"]
        self.collected: bool = multicompartment_data["collected"]

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class Operations:
    """Object representation of :class:`Parcel` `operations`

    :param operations_data: :class:`dict` containing operations data for :class:`Parcel`
    :type operations_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, operations_data: dict, logger: logging.Logger):
        """Constructor method

        :param operations_data: :class:`dict` containing operations data for :class:`Parcel`
        :type operations_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.manual_archive: bool = operations_data["manualArchive"]
        self.auto_archivable_since: arrow | None = (
            get(operations_data["autoArchivableSince"]) if "autoArchivableSince" in operations_data else None
        )
        self.delete: bool = operations_data.get("delete", None)
        self.pay_to_send: bool = operations_data.get("payToSend", None)
        self.collect: bool = operations_data.get("collect", None)
        self.expand_avizo: bool = operations_data.get("expandAvizo", None)
        self.highlight: bool = operations_data.get("highlight", None)
        self.refresh_until: arrow = get(operations_data["refreshUntil"]) if "refreshUntil" in operations_data else None
        self.request_easy_access_zone: str = operations_data.get("requestEasyAccessZone", None)
        self.is_voicebot: bool = operations_data.get("voicebot", None)
        self.can_share_to_observe: bool = operations_data.get("canShareToObserve", None)
        self.can_share_open_code: bool = operations_data.get("canShareOpenCode", None)
        self.can_share_parcel: bool = operations_data.get("canShareParcel", None)
        self.send: bool | None = operations_data.get("send", None)

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class EventLog:
    """Object representation of :class:`Parcel` single eventlog

    :param eventlog_data: :class:`dict` containing single eventlog data for :class:`Parcel`
    :type eventlog_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, eventlog_data: dict, logger: logging.Logger):
        """Constructor method

        :param eventlog_data: :class:`dict` containing single eventlog data for :class:`Parcel`
        :type eventlog_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        :raises UnknownStatusError: Unknown status in EventLog
        """

        self.type: str = eventlog_data["type"]
        self.date: arrow = get(eventlog_data["date"])
        self.details: dict | None = eventlog_data.get("details")
        self._log: logging.Logger = logger.getChild(self.__class__.__name__)

        if self.type == "PARCEL_STATUS":
            self.name = ParcelStatus[eventlog_data.get("name")]
        elif self.type == "RETURN_STATUS":
            self.name = ReturnsStatus[eventlog_data.get("name")]
        elif self.type == "PAYMENT":
            self.name = PaymentStatus[eventlog_data.get("name")]
        else:
            self._log.warning(f'Unknown status type {eventlog_data.get("name")}!')
            raise UnknownStatusError(reason=eventlog_data.get("name"))

        self._log.debug("created")

        if self.name == ParcelStatus.UNKNOWN or self.name == ReturnsStatus.UNKNOWN:
            self._log.warning(f'unknown {self.type}: {eventlog_data["name"]}')

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class SharedTo:
    """Object representation of :class:`Parcel` single shared to

    :param sharedto_data: :class:`dict` containing shared to data for :class:`Parcel`
    :type sharedto_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, sharedto_data: dict, logger: logging.Logger):
        """Constructor method

        :param sharedto_data: :class:`dict` containing shared to data for :class:`Parcel`
        :type sharedto_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.uuid: str = sharedto_data["uuid"]
        self.name: str = sharedto_data["name"]
        self.phone_number = sharedto_data["phoneNumber"]

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class QRCode:
    """Object representation of :class:`Parcel` QRCode

    :param qrcode_data: :class:`str` containing qrcode data for :class:`Parcel`
    :type qrcode_data: str
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, qrcode_data: str, logger: logging.Logger):
        """Constructor method

        :param qrcode_data: :class:`str` containing qrcode data for :class:`Parcel`
        :type qrcode_data: str
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self._qr_code = qrcode_data

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")

    @property
    def qr_image(self) -> BytesIO:
        """Returns a generated QR image for :class:`QRCode`

        :return: QR Code image
        :rtype: BytesIO"""

        self._log.debug("generating qr image")
        qr = qrcode.QRCode(
            version=3, error_correction=qrcode.constants.ERROR_CORRECT_H, box_size=20, border=4, mask_pattern=5
        )

        qr.add_data(self._qr_code)
        qr.make(fit=False)
        img1 = qr.make_image(fill_color="black", back_color="white")
        bio = BytesIO()
        bio.name = "qr.png"
        img1.save(bio, "PNG")
        bio.seek(0)
        self._log.debug("generated qr image")
        return bio


class CompartmentLocation:
    """Object representation of :class:`CompartmentProperties` compartment location

    :param compartmentlocation_data: :class:`dict` containing compartment location data for :class:`Parcel`
    :type compartmentlocation_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, compartmentlocation_data: dict, logger: logging.Logger):
        """Constructor method

        :param compartmentlocation_data: :class:`dict` containing compartment location data for :class:`Parcel`
        :type compartmentlocation_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.name: str = compartmentlocation_data["compartment"]["name"]
        self.side: str = compartmentlocation_data["compartment"]["location"]["side"]
        self.column: str = compartmentlocation_data["compartment"]["location"]["column"]
        self.row: str = compartmentlocation_data["compartment"]["location"]["row"]
        self.open_compartment_waiting_time: int = compartmentlocation_data["openCompartmentWaitingTime"]
        self.action_time: int = compartmentlocation_data["actionTime"]
        self.confirm_action_time: int = compartmentlocation_data["confirmActionTime"]

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")


class CompartmentProperties:
    """Object representation of :class:`Parcel` compartment properties

    :param compartmentproperties_data: :class:`dict` containing compartment properties data for :class:`Parcel`
    :type compartmentproperties_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger"""

    def __init__(self, compartmentproperties_data: dict, logger: logging.Logger):
        """Constructor method

        :param compartmentproperties_data: :class:`dict` containing compartment properties data for :class:`Parcel`
        :type compartmentproperties_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self._session_uuid: str = compartmentproperties_data["sessionUuid"]
        self._session_expiration_time: int = compartmentproperties_data["sessionExpirationTime"]
        self._location: CompartmentLocation | None = None
        self._status: CompartmentActualStatus | None = None

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")

    def __repr__(self):
        fields = tuple(f"{k}={v}" for k, v in self.__dict__.items() if k != "_log")
        return self.__class__.__name__ + str(tuple(sorted(fields))).replace("'", "")

    @property
    def session_uuid(self):
        """Returns a session unique identified for :class:`CompartmentProperties`

        :return: string containing session unique identified for :class:`CompartmentProperties`
        :rtype: str"""

        self._log.debug("getting session uuid")
        return self._session_uuid

    @property
    def location(self):
        """Returns a compartment location for :class:`CompartmentProperties`

        :return: compartment location for :class:`CompartmentProperties`
        :rtype: str
        """

        self._log.debug("getting location")
        return self._location

    @location.setter
    def location(self, location_data: dict):
        """Set a compartment location for :class:`CompartmentProperties`

        :param location_data: dict containing compartment location data for :class:`CompartmentProperties`
        :type location_data: dict
        """

        self._log.debug("setting location")
        self._location = CompartmentLocation(location_data, self._log)

    @property
    def status(self):
        """Returns a compartment status for :class:`CompartmentProperties`

        :return: compartment location for :class:`CompartmentProperties`
        :rtype: CompartmentActualStatus
        """

        self._log.debug("getting status")
        return self._status

    @status.setter
    def status(self, status_data: str | CompartmentActualStatus):
        self._log.debug("setting status")
        self._status = (
            status_data if isinstance(status_data, CompartmentActualStatus) else CompartmentActualStatus[status_data]
        )

        if self._status == CompartmentActualStatus.UNKNOWN and isinstance(status_data, str):
            self._log.warning(f"unexpected compartment actual status: {status_data}")


class AirSensorData:
    """Object representation of :class:`Parcel` air sensor data

    :param airsensor_data: :class:`dict` containing air sensor data for :class:`Parcel`
    :type airsensor_data: dict
    :param logger: :class:`logging.Logger` parent instance
    :type logger: logging.Logger
    """

    def __init__(self, airsensor_data: dict, logger: logging.Logger):
        """Constructor method

        :param airsensor_data: :class:`dict` containing air sensor data for :class:`Parcel`
        :type airsensor_data: dict
        :param logger: :class:`logging.Logger` parent instance
        :type logger: logging.Logger
        """

        self.updated_until: arrow = airsensor_data["updatedUntil"]
        self.air_quality: str = airsensor_data["airQuality"]
        self.temperature: float = airsensor_data["temperature"]
        self.humidity: float = airsensor_data["humidity"]
        self.pressure: float = airsensor_data["pressure"]
        self.pm25_value: float = airsensor_data["pollutants"]["pm25"]["value"]
        self.pm25_percent: float = airsensor_data["pollutants"]["pm25"]["percent"]
        self.pm10_value: float = airsensor_data["pollutants"]["pm10"]["value"]
        self.pm10_percent: float = airsensor_data["pollutants"]["pm10"]["percent"]

        self._log: logging.Logger = logger.getChild(self.__class__.__name__)
        self._log.debug("created")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inpost', 'inpost.static', 'static']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.4.0,<10.0.0',
 'aiohttp>=3.8.1,<4.0.0',
 'arrow>=1.2.3,<2.0.0',
 'pytest>=7.4.0,<8.0.0',
 'qrcode>=7.3.1,<8.0.0']

setup_kwargs = {
    'name': 'inpost',
    'version': '0.1.5',
    'description': 'Asynchronous InPost package allowing you to manage existing incoming parcels without mobile app',
    'long_description': "[![CodeFactor](https://www.codefactor.io/repository/github/ifossa/inpost-python/badge)](https://www.codefactor.io/repository/github/ifossa/inpost-python)\n![Code Quality](https://github.com/ifossa/inpost-python/actions/workflows/lint.yml/badge.svg?barnch=main)\n![Todos](https://github.com/ifossa/inpost-python/actions/workflows/todos.yml/badge.svg?barnch=main)\n\n# Inpost Python\n\nFully async Inpost library using Python 3.10.\n\n\n## Documentation\n\n[Readthedocs.io](https://inpost-python.readthedocs.io/en/latest/)\n\n\n## Usage/Examples\n\n\n```python\nfrom inpost.api import Inpost\n\ninp = Inpost('555333444')\nawait inp.send_sms_code():\n...\nif await inp.confirm_sms_code(123321):\n   print('Congratulations, you initialized successfully!')\n```\n\n## Before you contribute\n\n![Contributors](https://contrib.rocks/image?repo=ifossa/inpost-python)\n\nInstall linters and checkers, unlinted pull requests will not be approved\n```commandline\npoetry run pre-commit install\n```\n\n## Authors\n\n- [@loboda4450](https://www.github.com/loboda4450)\n- [@mrkazik99](https://www.github.com/mrkazik99)\n\n\n## Used By\n\nThis project is used by the following repos:\n\n[Inpost Telegram Bot](https://github.com/loboda4450/inpost-telegram-bot)\n\n\n\n## 😂 Here is a random joke that'll make you laugh!\n![Jokes Card](https://readme-jokes.vercel.app/api)",
    'author': 'loboda4450',
    'author_email': 'loboda4450@gmail.com',
    'maintainer': 'loboda4450',
    'maintainer_email': 'loboda4450@gmail.com',
    'url': 'https://github.com/IFOSSA/inpost-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
